﻿// (c) Copyright Microsoft Corporation.
// This source is subject to [###LICENSE_NAME###].
// Please see [###LICENSE_LINK###] for details.
// All other rights reserved.

using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Runtime.Serialization;
using System.Security.Permissions;

namespace System.Web.DomainServices
{
    // TODO: use these summary comments (or similar) when bug 597393 is addressed.
    ///// <summary>
    ///// This exception is used to indicate operational errors occurring during the
    ///// invocation of <see cref="DomainService"/> operations.
    ///// </summary>

    /// <summary>
    /// This exception is raised by generated <see cref="DomainService"/> proxies 
    /// when errors are encountered during invocation of <see cref="DomainService"/> 
    /// operations.
    /// </summary>
    /// <seealso cref="DomainService.CreateDomainServiceProxy&lt;TService&gt;()"/>
    [Serializable]
    [SuppressMessage("Microsoft.Design", "CA1032:ImplementStandardExceptionConstructors", Justification = "Associated EntityOperations as a non-null constructor parameter are required.")]
    public sealed class EntityOperationException : Exception
    {
        private IEnumerable<EntityOperation> _entityOperations;

        /// <summary>
        /// Initializes a new <see cref="EntityOperationException"/> instance.
        /// </summary>
        /// <param name="message">A message describing the exception.</param>
        /// <param name="entityOperation">The <see cref="EntityOperation"/> associated with this exception.</param>
        /// <exception cref="ArgumentNullException">if <paramref name="entityOperation"/> is null.</exception>
        public EntityOperationException(string message, EntityOperation entityOperation)
            : this(message, new[] { entityOperation })
        {
        }

        /// <summary>
        /// Constructor that takes serialization info
        /// </summary>
        /// <param name="info">The serialization info</param>
        /// <param name="context">The streaming context used for serialization</param>
        private EntityOperationException(SerializationInfo info, StreamingContext context)
            : base(info, context)
        {
            this._entityOperations = (IEnumerable<EntityOperation>)info.GetValue("EntityOperations", typeof(IEnumerable<EntityOperation>));
        }

        /// <summary>
        /// Initializes a new <see cref="EntityOperationException"/> instance.
        /// </summary>
        /// <param name="message">A message describing the exception.</param>
        /// <param name="entityOperations">The <see cref="EntityOperation"/>s associated with this exception.</param>
        /// <exception cref="ArgumentNullException">if <paramref name="entityOperations"/> is null.</exception>
        public EntityOperationException(string message, IEnumerable<EntityOperation> entityOperations)
            : base(message)
        {
            if (entityOperations == null)
            {
                throw new ArgumentNullException("entityOperations");
            }

            this._entityOperations = entityOperations.ToList().AsReadOnly();
        }

        /// <summary>
        /// Gets a collection of <see cref="EntityOperation"/> associated with this exception. 
        /// </summary>
        public IEnumerable<EntityOperation> EntityOperations
        {
            get
            {
                return this._entityOperations;
            }
        }

        /// <summary>
        /// ISerializable implementation.
        /// </summary>
        /// <param name="info">The serialization info</param>
        /// <param name="context">The streaming context used for serialization</param>
        [SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.SerializationFormatter)]
        public override void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            base.GetObjectData(info, context);
            info.AddValue("EntityOperations", this._entityOperations);
        }
    }
}