﻿// MetaEntityStorage.cs
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.IO.IsolatedStorage;
using System.Linq;
using System.Reflection;
using System.Runtime.Serialization;
using System.Windows.Ria.Data;

namespace System.Windows.Ria.Sync.Offline {

    internal static class MetaEntityStorage {

        private static EntityList GetEntityList(DomainContext domainContext, MetaEntity metaEntity) {
            Type entityType = metaEntity.CurrentEntity.GetType();

            // Get entitylist for entityType
            MethodInfo methodInfo = domainContext.Entities.GetType().GetMethod("GetEntityList", Type.EmptyTypes);
            Type[] genericArguments = new Type[] { entityType };
            MethodInfo genericMethodInfo = methodInfo.MakeGenericMethod(genericArguments);
            EntityList entityList = (EntityList)genericMethodInfo.Invoke(domainContext.Entities, null);

            return entityList;
        }

        // Gets list of entity types exposed in the domain context
        private static IEnumerable<Type> GetEntityTypes(Type domainContextType) {
            BindingFlags bindingFlags = BindingFlags.Instance | BindingFlags.Public;
            Type entityListType = typeof(EntityList);

            IEnumerable<PropertyInfo> entityListProperties =
                domainContextType.GetProperties(bindingFlags).Where(p => entityListType.IsAssignableFrom(p.PropertyType));
            IEnumerable<Type> entityTypes =
                entityListProperties.Select(p => p.PropertyType.GetGenericArguments()[0]);

            return entityTypes;
        }

        public static void Load(DomainContext domainContext, string fileName) {
            domainContext.Entities.Clear();

            string jsonData = Load(fileName);
            if (String.IsNullOrEmpty(jsonData)) {
                return;
            }

            IEnumerable<Type> entityTypes = GetEntityTypes(domainContext.GetType());
            List<List<MetaEntity>> metaEntities = MetaEntitySerializer.Deserialize(jsonData, entityTypes);

            // Add or Attach the entites back
            foreach (List<MetaEntity> list in metaEntities) {
                MetaEntity firstMetaEntity = list.FirstOrDefault();
                if (firstMetaEntity == null) {
                    continue;
                }

                EntityList entityList = GetEntityList(domainContext, firstMetaEntity);

                foreach (MetaEntity metaEntity in list) {
                    if (metaEntity.EntityState == EntityState.New) {
                        MethodInfo addMethod = entityList.GetType().GetMethod("Add");
                        addMethod.Invoke(entityList, new object[] { metaEntity.CurrentEntity });
                    }
                    else {
                        if (metaEntity.OriginalEntity != null)
                            entityList.Attach(metaEntity.OriginalEntity);
                        else
                            entityList.Attach(metaEntity.CurrentEntity);
                    }
                }
            }

            // Playback updates, domain method invocations and deletes
            foreach (List<MetaEntity> list in metaEntities) {
                MetaEntity firstMetaEntity = list.FirstOrDefault();
                if (firstMetaEntity == null) {
                    continue;
                }

                EntityList entityList = GetEntityList(domainContext, firstMetaEntity);

                foreach (MetaEntity metaEntity in list) {
                    switch (metaEntity.EntityState) {
                        case EntityState.Deleted:
                            // Play back entity deletes
                            Entity deletedEntity = entityList.Cast<Entity>().Single(x => x.GetIdentity().Equals(metaEntity.CurrentEntity.GetIdentity()));

                            MethodInfo removeMethod = entityList.GetType().GetMethod("Remove");
                            removeMethod.Invoke(entityList, new object[] { deletedEntity });
                            break;

                        case EntityState.Modified:
                            Entity activeEntity = metaEntity.CurrentEntity;

                            // Play back entity updates if any
                            if (metaEntity.OriginalEntity != null) {
                                UpdateEntity(metaEntity.CurrentEntity, metaEntity.OriginalEntity);
                                activeEntity = metaEntity.OriginalEntity;
                            }

                            // If there are domain method invocations, need to play that back too
                            if (metaEntity.EntityActions.Count() > 0) {
                                foreach (MetaEntityAction action in metaEntity.EntityActions) {
                                    MethodInfo method = activeEntity.GetType().GetMethod(action.Name);
                                    object[] methodParameters = action.HasParameters ? action.Parameters.ToArray() : null;

                                    method.Invoke(activeEntity, methodParameters);
                                }
                            }
                            break;
                    }
                }
            }
        }

        private static string Load(string fileName) {
            string data = String.Empty;

            using (IsolatedStorageFile isoFile = IsolatedStorageFile.GetUserStoreForApplication()) {
                if (isoFile.FileExists(fileName)) {
                    using (IsolatedStorageFileStream isoStream = new IsolatedStorageFileStream(fileName, FileMode.Open, isoFile)) {
                        using (StreamReader sr = new StreamReader(isoStream)) {
                            data = sr.ReadToEnd();
                        }
                    }
                }
            }
            return data;
        }

        public static void Save(DomainContext domainContext, string fileName) {
            IEnumerable<Type> entityTypes = GetEntityTypes(domainContext.GetType());
            EntityChangeSet changeSet = domainContext.Entities.GetChanges();

            List<List<MetaEntity>> metaEntities = new List<List<MetaEntity>>();

            foreach (Type entityType in entityTypes) {
                EntityList entityList = domainContext.Entities.GetEntityList(entityType);

                List<MetaEntity> metaEntityList = new List<MetaEntity>();
                foreach (Entity entity in entityList) {
                    Entity original = entity.GetOriginal();
                    MetaEntity offEntity = new MetaEntity(entity);
                    metaEntityList.Add(offEntity);
                }

                // Get deleted entities from changeset
                IEnumerable<Entity> removedEntites = changeSet.RemovedEntities.Where(e => e.GetType() == entityType);

                // Need to save deleted entities as well
                foreach (Entity entity in removedEntites) {
                    MetaEntity offEntity = new MetaEntity(entity);
                    metaEntityList.Add(offEntity);
                }

                metaEntities.Add(metaEntityList);
            }

            string jsonData = MetaEntitySerializer.Serialize(metaEntities, entityTypes);
            Save(jsonData, fileName);
        }

        private static void Save(string data, string fileName) {
            using (IsolatedStorageFile isoFile = IsolatedStorageFile.GetUserStoreForApplication()) {
                using (IsolatedStorageFileStream isoStream = new IsolatedStorageFileStream(fileName, FileMode.Create, isoFile)) {
                    using (StreamWriter sw = new StreamWriter(isoStream)) {
                        sw.Write(data);
                        sw.Flush();
                        sw.Close();
                    }
                }
            }
        }

        private static void UpdateEntity(Entity current, Entity original) {
            Type entityType = current.GetType();
            PropertyInfo[] props = entityType.GetProperties(BindingFlags.Public | BindingFlags.Instance);

            foreach (PropertyInfo prop in props) {
                object[] readOnly = prop.GetCustomAttributes(typeof(ReadOnlyAttribute), true);
                object[] dataMember = prop.GetCustomAttributes(typeof(DataMemberAttribute), true);

                if ((readOnly.Length == 0) && prop.CanWrite && (dataMember.Length > 0)) {
                    object newValue = prop.GetValue(current, null);
                    prop.SetValue(original, newValue, null);
                }
            }
        }
    }
}
