﻿// ServerSyncBehavior.cs
//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Net.NetworkInformation;
using System.Windows;
using System.Windows.Ria.Data;
using System.Windows.Ria.Sync.Offline;
using System.Windows.Threading;

namespace System.Windows.Ria.Sync {

    public sealed class ServerSyncBehavior<TDomainContext> : DomainSyncBehavior<TDomainContext> where TDomainContext : DomainContext {

        private Dispatcher _dispatcher;
        private bool _readOnly;
        private List<Func<TDomainContext, LoadOperation>> _loaders;

        private List<LoadOperation> _loadOperations;
        private SubmitOperation _submitOperation;

        public ServerSyncBehavior(TimeSpan syncInterval)
            : this(syncInterval, true) {
        }

        public ServerSyncBehavior(TimeSpan syncInterval, bool readOnly)
            : base(syncInterval) {
            _readOnly = readOnly;
            _loaders = new List<Func<TDomainContext, LoadOperation>>();
        }

        public void Load(Expression<Func<TDomainContext, LoadOperation>> loadAction) {
            Func<TDomainContext, LoadOperation> loader = loadAction.Compile();
            _loaders.Add(loader);
        }

        protected override void OnAttach() {
            NetworkChange.NetworkAddressChanged += OnNetworkAddressChanged;
            StartSync();
        }

        protected override void OnDetach() {
            NetworkChange.NetworkAddressChanged -= OnNetworkAddressChanged;
            StopSync();
        }

        private void OnNetworkAddressChanged(object sender, EventArgs e) {
            StartSync();
        }

        private void StartLoad() {
            _loadOperations = new List<LoadOperation>();
            foreach (Func<TDomainContext, LoadOperation> loader in _loaders) {
                LoadOperation loadOperation = loader(DomainContext);
                _loadOperations.Add(loadOperation);
            }
        }

        public override void StartSync() {
            bool isConnected = NetworkInterface.GetIsNetworkAvailable();
            if (isConnected) {
                base.StartSync();
            }
            else {
                base.StopSync();
            }
        }

        private void SubmitChangesCallback(SubmitOperation operation) {
            if (_submitOperation != operation) {
                return;
            }

            bool performLoad = false;
            if (operation.HasError == false) {
                performLoad = (bool)operation.UserState;
            }
            else {
                // TODO: Raise notification
            }

            _submitOperation = null;

            _dispatcher.BeginInvoke(delegate() {
                if (performLoad) {
                    StartLoad();
                }

                SyncingApplication.ExitSync();
                StartSync();
            });
        }

        protected internal override void Sync(bool performLoad, bool performSave) {
            if (DomainContext.IsLoading || DomainContext.IsSubmitting) {
                return;
            }

            bool isConnected = NetworkInterface.GetIsNetworkAvailable();
            if (isConnected == false) {
                StopSync();
                return;
            }

            if ((_readOnly == false) && DomainContext.HasChanges) {
                if (performSave) {
                    try {
                        _dispatcher = Application.Current.RootVisual.Dispatcher;

                        SyncingApplication.EnterSync();
                        _submitOperation = DomainContext.SubmitChanges(SubmitChangesCallback, performLoad);
                    }
                    catch {
                        SyncingApplication.ExitSync();
                    }
                }
            }
            else if (performLoad) {
                StartLoad();
            }
        }
    }
}
